"""Library for exceptions using the Portainer API."""


class PortainerException(Exception):
    """Base class for all client exceptions."""


class ApiException(PortainerException):
    """Raised during problems talking to the API."""


class AuthException(PortainerException):
    """Raised due to auth problems talking to API."""


class ApiBadRequestException(PortainerException):
    """Raised due sending a Rest command resulting in a bad request."""


class ApiForbiddenException(PortainerException):
    """Raised due to permission errors talking to API."""


class ApiUnauthorizedException(PortainerException):
    """Raised occasionally, mustn't harm the connection."""


class NoDataAvailableException(PortainerException):
    """Raised due updating data, when no data is available."""


class TimeoutException(PortainerException):
    """Raised due connecting the websocket."""



"""Models for Portainer API data."""

from dataclasses import dataclass, field
from enum import Enum
from typing import Any

from mashumaro import DataClassDictMixin, field_options


class Mode(Enum):
    """Enumeration for different modes."""

    EMPTY = ""
    RW = "rw"
    Z = "z"


class Propagation(Enum):
    """Enumeration for different propagation modes."""

    EMPTY = ""
    RPRIVATE = "rprivate"


class MountType(Enum):
    """Enumeration for different mount types."""

    BIND = "bind"
    VOLUME = "volume"


class PortType(Enum):
    """Enumeration for different port types."""

    TCP = "tcp"


class State(Enum):
    """Enumeration for container states."""

    CREATED = "created"
    DEAD = "dead"
    EXITED = "exited"
    RUNNING = "running"
    RESTARTING = "restarting"
    PAUSED = "paused"


@dataclass
class Container(DataClassDictMixin):
    """Dataclass for container information."""

    id: str = field(metadata=field_options(alias="Id"))
    name: str = field(
        metadata=field_options(
            alias="Names", deserialize=lambda x: x[0].strip("/").capitalize()
        )
    )
    image: str = field(metadata=field_options(alias="Image"))
    state: State = field(metadata=field_options(alias="State"))


@dataclass
class DockerSnapshotRaw(DataClassDictMixin):
    """Dataclass for raw Docker snapshot data."""

    containers: dict[str, Container] = field(
        metadata=field_options(
            alias="Containers",
            deserialize=lambda containers_list: {
                container_data["Id"]: Container.from_dict(container_data)
                for container_data in containers_list
            },
        )
    )


@dataclass
class Snapshot(DataClassDictMixin):
    """Dataclass for snapshot information."""

    time: int = field(metadata=field_options(alias="Time"))
    docker_version: str = field(metadata=field_options(alias="DockerVersion"))
    swarm: bool = field(metadata=field_options(alias="Swarm"))
    total_cpu: int = field(metadata=field_options(alias="TotalCPU"))
    total_memory: int = field(metadata=field_options(alias="TotalMemory"))
    running_container_count: int = field(
        metadata=field_options(alias="RunningContainerCount")
    )
    stopped_container_count: int = field(
        metadata=field_options(alias="StoppedContainerCount")
    )
    healthy_container_count: int = field(
        metadata=field_options(alias="HealthyContainerCount")
    )
    unhealthy_container_count: int = field(
        metadata=field_options(alias="UnhealthyContainerCount")
    )
    volume_count: int = field(metadata=field_options(alias="VolumeCount"))
    image_count: int = field(metadata=field_options(alias="ImageCount"))
    service_count: int = field(metadata=field_options(alias="ServiceCount"))
    stack_count: int = field(metadata=field_options(alias="StackCount"))
    docker_snapshot_raw: DockerSnapshotRaw = field(
        metadata=field_options(alias="DockerSnapshotRaw")
    )
    node_count: int = field(metadata=field_options(alias="NodeCount"))
    gpu_use_all: bool = field(metadata=field_options(alias="GpuUseAll"))
    gpu_use_list: list[Any] = field(metadata=field_options(alias="GpuUseList"))


@dataclass
class TLSConfig(DataClassDictMixin):
    """Dataclass for TLS configuration."""

    tls: bool = field(metadata=field_options(alias="TLS"))
    tls_skip_verify: bool = field(metadata=field_options(alias="TLSSkipVerify"))


@dataclass
class NodeData(DataClassDictMixin):
    """Dataclass for node data."""

    id: int = field(metadata=field_options(alias="Id"))
    name: str = field(metadata=field_options(alias="Name"))
    type: int = field(metadata=field_options(alias="Type"))
    url: str = field(metadata=field_options(alias="URL"))
    group_id: int = field(metadata=field_options(alias="GroupId"))
    public_url: str = field(metadata=field_options(alias="PublicURL"))
    gpus: list[Any] = field(metadata=field_options(alias="Gpus"))
    tls_config: TLSConfig = field(metadata=field_options(alias="TLSConfig"))
    status: int = field(metadata=field_options(alias="Status"))
    snapshots: list[Snapshot] = field(metadata=field_options(alias="Snapshots"))
    enable_gpu_management: bool = field(
        metadata=field_options(alias="EnableGPUManagement")
    )
    authorized_users: str = field(metadata=field_options(alias="AuthorizedUsers"))
    authorized_teams: str = field(metadata=field_options(alias="AuthorizedTeams"))
    tags: str = field(metadata=field_options(alias="Tags"))
    is_edge_device: bool = field(metadata=field_options(alias="IsEdgeDevice"))


@dataclass
class PortainerList(DataClassDictMixin):
    """DataClass for a list of all portainers."""

    data: list[NodeData]



"""Module to connect to the portainer server."""

import asyncio
import contextlib
import logging
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any

from .auth import AbstractAuth
from .const import REST_POLL_CYCLE
from .model import NodeData
from .portainer_settings import PortainerSettings
from .utils import portainer_list_to_dictionary

_LOGGER = logging.getLogger(__name__)

logging.basicConfig(level=logging.DEBUG)


@dataclass
class PortainerEndpoint:
    """Endpoint URLs for the Portainer API."""

    settings = "api/settings"
    "List data for all portainer instances."

    endpoints = "api/endpoints"
    "List data for all portainer instances."

    endpoints_env = "api/endpoints/{env_id}"
    "List data for a specific env_id."

    restart = "api/endpoints/{env_id}/docker/containers/{container_id}/restart"
    "Restart a specific container in an environment."

    start = "api/endpoints/{env_id}/docker/containers/{container_id}/start"
    "Start a specific container in an environment."

    stop = "api/endpoints/{env_id}/docker/containers/{container_id}/stop"
    "Stop a specific container in an environment."


class PortainerClient:
    """API to communicate with an Portainer.

    The `PortainerClient` is the primary API service for this library. It supports
    operations like getting a status or sending commands.
    """

    def __init__(
        self,
        auth: AbstractAuth,
        poll: bool = False,
    ) -> None:
        """Create a client.

        :param class auth: The AbstractAuth class from aiotainer.auth.
        :param bool poll: Poll data with rest if True.
        """
        self._data: dict[str, Iterable[Any]] | None = {}
        self.auth = auth
        self.data: dict[int, NodeData] = {}
        self.loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
        self.poll = poll
        self.rest_task: asyncio.Task | None = None

    async def connect(self) -> None:
        """Connect to the API.

        This method handles the login. Also a REST task will be started, which
        periodically polls the REST endpoint, when polling is set to true.
        """

        if self.poll:
            await self.get_status()
            self.rest_task = asyncio.create_task(self._rest_task())

    async def get_settings(self) -> PortainerSettings:
        """Get status of all endpoints."""
        portainer_list = await self.auth.get_json(PortainerEndpoint.settings)
        return PortainerSettings.from_dict(portainer_list)

    async def get_status(self) -> dict[int, NodeData]:
        """Get status of all endpoints."""
        portainer_list = await self.auth.get_json(PortainerEndpoint.endpoints)
        self.data = portainer_list_to_dictionary(portainer_list)
        return self.data

    async def get_status_specific(self, env_id: int) -> NodeData:
        """Get status of a specific endpoint."""
        portainer_list = await self.auth.get_json_node(
            PortainerEndpoint.endpoints_env.format(env_id=env_id)
        )
        self.data[env_id] = NodeData.from_dict(portainer_list)
        return self.data[env_id]

    async def restart_container(self, env_id: int, container_id: str):
        """Restart container."""
        url = PortainerEndpoint.restart.format(env_id=env_id, container_id=container_id)
        await self.auth.post(url)

    async def start_container(self, env_id: int, container_id: str):
        """Start container."""
        url = PortainerEndpoint.start.format(env_id=env_id, container_id=container_id)
        await self.auth.post(url)

    async def stop_container(self, env_id: int, container_id: str):
        """Stop container."""
        url = PortainerEndpoint.stop.format(env_id=env_id, container_id=container_id)
        await self.auth.post(url)

    async def _rest_task(self) -> None:
        """Poll data periodically via Rest."""
        while True:
            await self.get_status()
            await asyncio.sleep(REST_POLL_CYCLE)

    async def close(self) -> None:
        """Close the client."""
        if self.rest_task:
            if not self.rest_task.cancelled():
                self.rest_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await asyncio.gather(self.rest_task)
